/* pwd - print working directory	Author: Norbert Schlenker */

/*
 * pwd - print working directory
 *   Syntax:	pwd
 *   Flags:	None.
 *   Author:	Norbert Schlenker
 *   Copyright:	None.  Released to the public domain.
 *   Reference:	IEEE P1003.2 Section 4.50 (draft 10)
 *   Bugs:	No internationalization support; all messages are in English.
 */

/* Force visible Posix names */
#ifndef _POSIX_SOURCE
#define _POSIX_SOURCE 199506L
#define _REENTRANT
#endif

/* External interfaces */
#include <sys/types.h>
#include <dirent.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/stat.h>
#include <time.h>

/* Magic numbers suggested or required by Posix specification */
#define SUCCESS	0		/* exit code in case of success */
#define FAILURE 1		/*                   or failure */

static char dir[PATH_MAX + 1];

int pwd()
{
  char *d = getcwd(dir, PATH_MAX);
  printf("%s\n", d);
  return d?SUCCESS:FAILURE;
}



int cd(char *path){
  int rval = chdir(path);
  pwd();
  return rval;
}

#define FOUND 1
#define NOT_FOUND 0

/* lifted from minix ls 5.0 - List files. Author: Kees J. Bot
 *					   		25 Apr 1989
 *
 */

#define	PAST	(26*7*24*3600L)	/* Half a year ago. */
/* Between PAST and FUTURE from now a time is printed, otherwise a year. */
#define FUTURE	( 1*7*24*3600L)	/* One week. */

static char *timestamp(char *date, time_t t)
/* Transform the right time field into something readable. */
{
  struct tm res;
	struct tm *tm;
	time_t now;
	static char month[] = "JanFebMarAprMayJunJulAugSepOctNovDec";

	tm= localtime_r(&t,&res);
	time(&now);

	if (t < now - PAST || t > now + FUTURE) {
		sprintf(date, "%.3s %2d  %d",
			month + 3*tm->tm_mon,
			tm->tm_mday,
			1900 + tm->tm_year);
	} else {
		sprintf(date, "%.3s %2d %02d:%02d",
			month + 3*tm->tm_mon,
			tm->tm_mday,
			tm->tm_hour, tm->tm_min);
	}
	return date;
}

/* The array l_ifmt[] is used in an 'ls -l' to map the type of a file to a
 * letter.  This is done so that ls can list any future file or device type
 * other than symlinks, without recompilation.  (Yes it's dirty.)
 */
static const char l_ifmt[] = "0pcCd?bB-?l?s???";

#define ifmt(mode)	l_ifmt[((mode) >> 12) & 0xF]

static char *permissions(char *rwx, struct stat *f)
/* Compute long or short rwx bits. */
{
  char *p= rwx+1;
  int mode= f->st_mode;

  rwx[0] = ifmt(f->st_mode);
  /* Note that rwx[0] is a guess for the more alien file types.  It is
   * correct for BSD4.3 and derived systems.  I just don't know how
   * "standardized" these numbers are.
   */
  
  do {
    p[0] = (mode & S_IRUSR) ? 'r' : '-';
    p[1] = (mode & S_IWUSR) ? 'w' : '-';
    p[2] = (mode & S_IXUSR) ? 'x' : '-';
    mode<<= 3;
  } while ((p+=3) <= rwx+7);
  
  if (f->st_mode&S_ISUID) rwx[3]= f->st_mode&(S_IXUSR>>0) ? 's' : '=';
  if (f->st_mode&S_ISGID) rwx[6]= f->st_mode&(S_IXUSR>>3) ? 's' : '=';
  
  return rwx;
}

static int list_file(char *path, char *filename, int long_listing){
  char rwx[11]; /*  = "drwxr-x--x"; */
  char date[21]; /* = "Jan 19 03:14:07 2038"; */
  struct stat buf;
  char n[1024];
  char tbuf[20];
  
  sprintf(n,"%s/%s", path, filename);
  if (stat(n,&buf)==0){
    if(long_listing){
      printf("%s %-9d %s %s\n",
	     permissions(rwx, &buf),
	     buf.st_size,
	     timestamp(date, buf.st_mtime),
	     filename);
    } else {
      printf("%s\n", filename);
    } 
  } else {
    printf("Unable to stat file \"%s\"\n", n);
    return NOT_FOUND;
  }
  return FOUND;
}

static int list_dir(char *path, int long_listing){
  DIR *dirp;
  struct dirent *dp;

  dirp = opendir(path);
  while (dp = readdir(dirp)){
      list_file(path, dp->d_name, long_listing);
  }
  closedir(dirp);
  return NOT_FOUND;
}

int ls(char *path){
  list_dir(path?path:".",0);
}

int ll(char *path){
  list_dir(path?path:".",1);
}
