#ifdef POSIX_THREAD
#  include <pthread.h>
#  include <signal.h>
#  define PID_T pthread_t
#endif
#ifdef rtems
#  include <rtems.h>
#  define PID_T Objects_Id
#endif
#include <stdarg.h>
#include "libstab.h"

#define MAX_ARGS 10

/* must have MAX_ARGS arguments */
typedef int (*ff_t)
     (void *,void *,void *,void *,void *,
      void *,void *,void *,void *,void *);

#define ffcall(p) \
  (((ff_t)p)(arglist[0],arglist[1],arglist[2],arglist[3],arglist[4],\
             arglist[5],arglist[6],arglist[7],arglist[8],arglist[9]))


#define reset_args() (memset(arglist, 0, sizeof(arglist)),curr_arg = 0)
#define add_arg(arg) ((curr_arg <= MAX_ARGS)?arglist[curr_arg++]=(arg):0)

static void *arglist[MAX_ARGS+1];

static void trampoline(){
  ((ff_t)arglist[MAX_ARGS])
    (arglist[0],arglist[1],arglist[2],arglist[3],arglist[4],\
     arglist[5],arglist[6],arglist[7],arglist[8],arglist[9]);
}

/* 
 * Spawn a new task.  There may be many arguments (up to ten)so this
 * routine packages the argument list into an array passed as a global
 * variable to trampoline(), which invokes the function with the
 * arguments.
 */
/*VARARGS1*/
int sp(void *entry, ...)
{
  va_list ap;
  int curr_arg = 0;
  PID_T tid = -1;
#ifdef rtems
  static int last_id = 0;
  rtems_name name;
  rtems_status_code result;
#endif

  va_start(ap, entry);
  if(entry==0)
    return -1;

  memset(arglist,0,sizeof(arglist));
  arglist[MAX_ARGS]=entry;

  while(curr_arg < MAX_ARGS){
    arglist[curr_arg] = va_arg(ap, void *);
    curr_arg++;
  }
  va_end(ap);

#ifdef POSIX_THREADS
  pthread_create(&tid, 0, (void * (*)(void *))trampoline, 0);
#else
#  ifdef rtems
  last_id=(last_id+1)%1000;
  name = rtems_build_name('t', 
			  last_id/100 + 48, 
			  last_id/10  + 48,
			  last_id     + 48);
  result = rtems_task_create(name,
			     10,     /* priority */
			     RTEMS_MINIMUM_STACK_SIZE,
			     RTEMS_DEFAULT_ATTRIBUTES,
			     RTEMS_LOCAL,
			     &tid);
  rtems_task_start(tid,
		   (rtems_task_entry) trampoline,
		   (rtems_task_argument) 0);
#  endif
#endif
  return tid;
}

/* task delete */
int td(int tid){
#ifdef POSIX_THREADS
  return pthread_cancel(tid);
#else
#  ifdef rtems
  return rtems_task_delete(tid);
#  endif
#endif
}

/* task suspend */
int ts(int tid){
#ifdef POSIX_THREADS
  return pthread_kill(tid, SIGSTOP);
#else
#  ifdef rtems
  return rtems_task_suspend(tid);
#  endif
#endif
}

/* task resume */
int tr(int tid){
#ifdef POSIX_THREADS
  return pthread_kill(tid, SIGCONT);
#else
#  ifdef rtems
  return rtems_task_resume(tid);
#  endif
#endif
}

/* step next instruction */
int s(int tid){
  /* NOT IMPLEMENTED */
  return tid;
}

/* task status */
int i(void){
  /* NOT IMPLEMENTED */
  return 0;
}

/* This file uses gcc-specific calls to access the stack frame.
 */

#if ! defined(__GNUC__)
#  error This file can be compiled with GCC only!
#endif

static char *get_frame_address(int depth){
  /* looks stupid, but __builtin_frame_address 
   * accepts only a constant int as an arg.
   */
  
  switch(depth){
  case 1:  return __builtin_frame_address(1);
  case 2:  return __builtin_frame_address(2);
  case 3:  return __builtin_frame_address(3);
  case 4:  return __builtin_frame_address(4);
  case 5:  return __builtin_frame_address(5);
  case 6:  return __builtin_frame_address(6);
  case 7:  return __builtin_frame_address(7);
  case 8:  return __builtin_frame_address(8);
  case 9:  return __builtin_frame_address(9);
  case 10: return __builtin_frame_address(10);
  case 11: return __builtin_frame_address(11);
  case 12: return __builtin_frame_address(12);
  case 13: return __builtin_frame_address(13);
  case 14: return __builtin_frame_address(14);
  case 15: return __builtin_frame_address(15);
  case 16: return __builtin_frame_address(16);
  case 17: return __builtin_frame_address(17);
  case 18: return __builtin_frame_address(18);
  case 19: return __builtin_frame_address(19);
  case 20: return __builtin_frame_address(20);
  default: return 0;
  }
}

static char *get_return_address(int depth){
  /* looks stupid, but __builtin_return_address 
   * accepts only a constant int as an arg.
   */
  
  switch(depth){
  case 1:  return __builtin_return_address(1);
  case 2:  return __builtin_return_address(2);
  case 3:  return __builtin_return_address(3);
  case 4:  return __builtin_return_address(4);
  case 5:  return __builtin_return_address(5);
  case 6:  return __builtin_return_address(6);
  case 7:  return __builtin_return_address(7);
  case 8:  return __builtin_return_address(8);
  case 9:  return __builtin_return_address(9);
  case 10: return __builtin_return_address(10);
  case 11: return __builtin_return_address(11);
  case 12: return __builtin_return_address(12);
  case 13: return __builtin_return_address(13);
  case 14: return __builtin_return_address(14);
  case 15: return __builtin_return_address(15);
  case 16: return __builtin_return_address(16);
  case 17: return __builtin_return_address(17);
  case 18: return __builtin_return_address(18);
  case 19: return __builtin_return_address(19);
  case 20: return __builtin_return_address(20);
  default: return 0;
  }
}


/* This routine prints the task stack trace of the current program.
 * In RTEMS, we really want to walk the call stack of *any*
 * task. Perhaps someone could implement this properly.
 * Someone needs to create a set of routines (that are hopefully
 * not cpu-specific - a pipe dream perhaps) that can walk the stack
 * frame of any task to replace get_frame_address() and 
 * get_return_address().
 */

int tt(int tid, int depth) {
  void *frame_addr;
  void *return_addr;
  int i;
  int d = 1;
  stab_entry *sym;
  char *addr;

  /* if someone enters no depth, then default to 20: */
  depth = depth?depth:20;

  printf("Call stack:\n");
  while(depth>0){
    frame_addr = get_frame_address(d);
    if (!frame_addr) 
      break;
    return_addr = get_return_address(d);
    if(sym = stab_find_by_addr(system_symbol_table,return_addr)){
      addr = (char *)return_addr - (unsigned long)sym->address;
      printf("\t<%s+%d>\n", sym->name, addr);
    }else{
      printf("\t0x%08lx\n", return_addr);
    }

    depth--;
    d++;
  }
}
